/* $Id: bpfFilter.h,v 1.10.2.1 2003/03/19 02:57:18 lergnom Exp $ -*- C++ -*-
 * 
 * libbpfl - Bayesian Probability Filtering Library
 * Copyright (C) 2003 Anthony Ventimiglia
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 */
#ifndef BPF_FILTER_H
#define BPF_FILTER_H

#ifdef __BPF_BUILD

#include "common.h"
#include "bpfHash_map.h"
#include "bpfBucket.h"

#else  // __BPF_BUILD

#include <libbpfl/bpfHash_map.h>
#include <libbpfl/bpfBucket.h>

#endif // __BPF_BUILD

#include <vector>
#include <cmath>
#include <istream>

namespace bpf {

///This is used to do the actual fitering functions. It must be passed at
///least two bpfHash<> objects.
template <typename TokenType, typename HashObject>
class Filter
{
public:
    typedef Bucket<TokenType, HashObject> bucket_type;

    Filter () {}
    virtual ~Filter () {}

    typedef struct score_info;

    virtual void add_bucket ( bucket_type* );
    ///Find score for one token, one score for each bucket.
    virtual std::vector<double> score_token ( const TokenType& );
    /// Returns array of bucket names
    virtual std::vector<std::string> get_bucket_names ();
    /// Calculates score for every token in stream.
    /// Returns all the scores, as well as the list of tokens read in
    /// the score_info return struct.
    
    ///This is used to find the scores of a stream of tokens.
    score_info* score_stream ( std::istream& );
    
private:
    std::vector<bucket_type*> buckets;
};

// Include score_info template here.
#ifdef __BPF_BUILD
#include "Filter_score_info.h"
#else
#include <libbpfl/Filter_score_info.h>
#endif

template <typename Tok, typename Hash>
typename Filter<Tok, Hash>::score_info*
Filter<Tok, Hash>::score_stream ( std::istream& istr )
{
    using std::vector;
    score_info *info = new score_info();
    
    // Read tokens, store them in info
    while ( istr ) {
	Tok token;
	istr >> token;
	if ( istr.eof() ) break;
	info->tokens.push_back ( token );
    }

    // First set up scores matrix for scores
    info->scores.resize ( buckets.size() ); // Set first dimension
    for ( size_t i = 0; i < info->scores.size(); ++i ) // And second
	info->scores[i].resize ( info->tokens.size() );

    // Calculate scores for all tokens in all buckets.
    for ( size_t tokInd = 0; tokInd < info->tokens.size(); ++tokInd ) {
	vector<double> tokenScores = score_token ( info->tokens[tokInd] );
	
	for ( size_t bucketInd = 0; bucketInd < buckets.size(); ++bucketInd )
	    info->scores[bucketInd][tokInd] = tokenScores[bucketInd];
    
    }
    
    // Now calculate each bucket's score
    info->bucket_scores.resize ( buckets.size() );
    double sumOfScores = 0;
    for ( size_t bucket = 0; bucket < buckets.size(); ++bucket ) {
	info->bucket_scores[bucket] = 1.0;
	for ( size_t tok = 0; tok < info->tokens.size(); ++tok )
	    info->bucket_scores[bucket] *= info->scores[bucket][tok];
	sumOfScores += info->bucket_scores[bucket];
    }
    for ( size_t b = 0; b < buckets.size(); ++b )
	info->bucket_scores[b] /= sumOfScores;
	    
	
    return info;
}

template <typename Tok, typename Hash>
void Filter<Tok, Hash>::add_bucket ( Filter::bucket_type* b )
{
    buckets.push_back ( b );
}

template <typename Tok, typename Hash>
std::vector<double> Filter<Tok, Hash>::score_token ( const Tok& t )
{
    std::vector<double> scores;
    scores.resize ( buckets.size() );
    double sum_of_scores = 0;
    for ( size_t i = 0; i < buckets.size(); ++i ) {
	scores[i] = buckets[i]->score ( t );
	sum_of_scores += scores[i];
    }
    for ( size_t i = 0; i < scores.size(); ++i ) {
	scores[i] /= sum_of_scores;
	if ( std::isnan ( scores[i] ) )
	    scores[i] = 0.00001;
	scores[i] = scores[i] < 0.00001 ? 0.00001 : scores[i];
	scores[i] = scores[i] > 0.99999 ? 0.99999 : scores[i];
    }
    return scores;
}

template < typename Tok, typename Hash>
std::vector<std::string> 
Filter<Tok, Hash>::get_bucket_names ()
{
    using namespace std;
    vector<string> names;
    names.resize ( buckets.size() );
    for ( size_t i = 0; i < buckets.size(); ++i )
	names[i] = buckets[i]->name;
    return names;
}


};				// namespace bpf
#endif // BPF_FILTER_H
